@@ -21,9 +21,6 @@
  ****************************************************************
  */
 
-#include "rcs.h"
-RCS_ID("$Id: ansi.c,v 1.22 1994/05/31 12:31:25 mlschroe Exp $ FAU")
-
 #include <sys/types.h>
 #include <fcntl.h>
 #ifndef sun	/* we want to know about TIOCPKT. */
@@ -37,6 +34,8 @@ RCS_ID("$Id: ansi.c,v 1.22 1994/05/31 12:31:25 mlschroe Exp $ FAU")
 #include "logfile.h"
 
 extern struct display *display, *displays;
+extern struct win *fore;	/* for 83 escape */
+extern struct layer *flayer;	/* for 83 escape */
 
 extern struct NewWindow nwin_default;	/* for ResetWindow() */
 extern int  nversion;		/* numerical version of screen */
@@ -284,10 +283,6 @@ char *s;
  *  - record program output in window scrollback
  *  - translate program output for the display and put it into the obuf.
  *
- * Output is only supressed, where obuf is beyond maximum and the flag
- * nonblock is set. Then we set nonblock from 1 to 2 and output a '~'
- * character instead. nonblock should be reset to 1 by a successfull
- * write.  Where nonblock isn't set, the obufmax is ignored.
  */
 void
 WriteString(wp, buf, len)
@@ -311,23 +306,6 @@ register int len;
   cols = curr->w_width;
   rows = curr->w_height;
 
-  /* The status should be already gone, so this is "Just in Case" */
-  for (cv = wp->w_layer.l_cvlist; cv; cv = cv->c_lnext)
-    {
-      display = cv->c_display;
-#if 0	/* done by new status code */
-      if (D_status == STATUS_ON_WIN)
-	RemoveStatus();
-#endif
-      if (D_nonblock == 1 && (D_obufp - D_obuf > D_obufmax))
-	{
-	  /* one last surprising '~' means: lost data */
-	  AddChar('~');
-	  /* private flag that prevents more output */
-	  D_nonblock = 2;
-	}
-    }
-
   if (curr->w_silence)
     SetTimeout(&curr->w_silenceev, curr->w_silencewait * 1000);
 
@@ -754,7 +732,7 @@ register int len;
 		}
 	    }
 #  endif
-	  if (font == 031 && c == 0x80)
+	  if (font == 031 && c == 0x80 && !curr->w_mbcs)
 	    font = curr->w_rend.font = 0;
 	  if (is_dw_font(font) && c == ' ')
 	    font = curr->w_rend.font = 0;
@@ -835,10 +813,14 @@ register int len;
 	      if (c == 0x80 && font == 0 && curr->w_encoding == GBK)
 		c = 0xa4;
 	      else
-#endif
 	        c &= 0x7f;
+	      if (c < ' ' && font != 031)
+		goto tryagain;
+#else
+	      c &= 0x7f;
 	      if (c < ' ')	/* this is ugly but kanji support */
 		goto tryagain;	/* prevents nicer programming */
+#endif
 	    }
 #endif /* FONT */
 	  if (c == '\177')
@@ -1485,19 +1467,26 @@ StringEnd()
 	{
 	  /* special execute commands sequence */
 	  char *args[MAXARGS];
+	  int argl[MAXARGS];
 	  struct acluser *windowuser;
 
 	  windowuser = *FindUserPtr(":window:");
-	  if (windowuser && Parse(p, args))
+	  if (windowuser && Parse(p, sizeof(curr->w_string) - (p - curr->w_string), args, argl))
 	    {
 	      for (display = displays; display; display = display->d_next)
 		if (D_forecv->c_layer->l_bottom == &curr->w_layer)
 		  break;	/* found it */
 	      if (display == 0 && curr->w_layer.l_cvlist)
 		display = curr->w_layer.l_cvlist->c_display;
+	      if (display == 0)
+		display = displays;
 	      EffectiveAclUser = windowuser;
-	      DoCommand(args);
+	      fore = curr;
+	      flayer = fore->w_savelayer ? fore->w_savelayer : &fore->w_layer;
+	      DoCommand(args, argl);
 	      EffectiveAclUser = 0;
+	      fore = 0;
+	      flayer = 0;
 	    }
 	  break;
 	}
@@ -1566,7 +1555,7 @@ StringEnd()
     case AKA:
       if (curr->w_title == curr->w_akabuf && !*curr->w_string)
 	break;
-      ChangeAKA(curr, curr->w_string, 20);
+      ChangeAKA(curr, curr->w_string, strlen(curr->w_string));
       if (!*curr->w_string)
 	curr->w_autoaka = curr->w_y + 1;
       break;
@@ -1582,9 +1571,11 @@ PrintStart()
   curr->w_pdisplay = 0;
 
   /* find us a nice display to print on, fore prefered */
-  for (display = displays; display; display = display->d_next)
-    if (curr == D_fore && (printcmd || D_PO))
-      break;
+  display = curr->w_lastdisp;
+  if (!(display && curr == D_fore && (printcmd || D_PO)))
+    for (display = displays; display; display = display->d_next)
+      if (curr == D_fore && (printcmd || D_PO))
+        break;
   if (!display)
     {
       struct canvas *cv;
@@ -2131,7 +2122,7 @@ SelectRendition()
 	cx &= 0x0f;
 # endif
 #endif
-      if (j < 0 || j >= (sizeof(rendlist)/sizeof(*rendlist)))
+      if (j < 0 || j >= (int)(sizeof(rendlist)/sizeof(*rendlist)))
 	continue;
       j = rendlist[j];
       if (j & (1 << NATTR))
@@ -2184,8 +2175,10 @@ int l;
 {
   int i, c;
 
-  for (i = 0; i < 20 && l > 0; l--)
+  for (i = 0; l > 0; l--)
     {
+      if (p->w_akachange + i == p->w_akabuf + sizeof(p->w_akabuf) - 1)
+	break;
       c = (unsigned char)*s++;
       if (c == 0)
 	break;
@@ -2636,6 +2629,7 @@ int x, y;
   MFixLine(p, y, c);
   ml = &p->w_mlines[y];
   MKillDwRight(p, ml, x);
+  MKillDwLeft(p, ml, x);
   copy_mchar2mline(c, ml, x);
 #ifdef DW_CHARS
   if (c->mbcs)
@@ -3060,3 +3054,4 @@ int what;
     }
   display = olddisplay;
 }
+
